#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/pid.h>
#include <linux/pgtable.h>
#include <linux/moduleparam.h>


MODULE_DESCRIPTION("LKM3");
MODULE_AUTHOR("CS695");
MODULE_LICENSE("GPL");

static int pid = 1;
module_param(pid, int, 0);
MODULE_PARM_DESC(pid, "PID of the process");

static unsigned long long int vaddr = 0;
module_param(vaddr, ullong, 0);
MODULE_PARM_DESC(vaddr, "Virtual address of the process");


static __init int print_physical_address(void)
{
	struct task_struct *task = NULL;
	struct mm_struct *mm;
	pgd_t* pgd;
	p4d_t* p4d;
	pud_t* pud;
	pmd_t* pmd;
	pte_t* pte;

	unsigned long long int phy_addr;

	rcu_read_lock();

	task = pid_task(find_vpid(pid), PIDTYPE_PID);
	
	if (task == NULL) {
		pr_info("[LKM3] Process with PID: %d not found\n", pid);
		rcu_read_unlock();
		return 0;
	}

	mm = task->mm;
	// Page table walk
	pgd = pgd_offset(mm, vaddr);
	if(pgd_none(*pgd) || pgd_bad(*pgd)) {
		pr_info("[LKM3] Page not present\n");
		return 0;
	}

	p4d = p4d_offset(pgd, vaddr);
	if(p4d_none(*p4d) || p4d_bad(*p4d)) {
		pr_info("[LKM3] Page not present\n");
		return 0;
	}
	
	pud = pud_offset(p4d, vaddr);
	if(pud_none(*pud) || pud_bad(*pud)) {
		pr_info("[LKM3] Page not present\n");
		return 0;
	}

	pmd = pmd_offset(pud, vaddr);
	if(pmd_none(*pmd) || pmd_bad(*pmd)) {
		pr_info("[LKM3] Page not present\n");
		return 0;
	}

	pte = pte_offset_kernel(pmd, vaddr);
	if(pte_none(*pte)) {
		pr_info("[LKM3] Page not present\n");
		return 0;
	}

	phy_addr = (pte_pfn(*pte) << PAGE_SHIFT) | (vaddr & ~PAGE_MASK); 		// works on both x86_64 and arm64
	// phy_addr = (pte_val(*pte) & PTE_PFN_MASK) | (vaddr & ~PAGE_MASK); 	// alternative, works on x86_64
	// phy_addr = (pte_val(*pte) & PTE_ADDR_MASK) | (vaddr & ~PAGE_MASK); 	// alternative, works on arm64
	// phy_addr = (__pte_to_phys(*pte)) | (vaddr & ~PAGE_MASK); 			// alternative, works on arm64

	rcu_read_unlock();

	pr_info("[LKM3] Virtual address: 0x%llx / %llu\n",vaddr, vaddr);
	pr_info("[LKM3] Physical address: 0x%llx / %llu\n",phy_addr, phy_addr);

	return 0;

}

static __exit void exit_module(void)
{
	pr_info("[LKM3] Module LKM3 Unloaded\n");
}

module_init(print_physical_address);
module_exit(exit_module);
