#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/sched/signal.h>
#include <linux/sched.h>
#include <linux/mm_types.h>
#include <linux/pgtable.h>
#include <asm/io.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/errno.h>
#include <linux/highmem.h>

#include "chardev.h"

#define FIRST_MINOR 0
#define MINOR_CNT 1

static dev_t dev;
static struct cdev c_dev;
static struct class *cl;

static unsigned long page_table_walk(unsigned long virt_addr,struct mm_struct* mm)
{
	pgd_t* pgd;
	p4d_t* p4d;
	pud_t* pud;
	pmd_t* pmd;
	pte_t* pte;
	unsigned long long int phy_addr;

	// Page table walk
	pgd = pgd_offset(mm, virt_addr);
	if(pgd_none(*pgd) || pgd_bad(*pgd)){
		pr_info("[CHAR_DEV] Page not present\n");
		return 0;
	}

	p4d = p4d_offset(pgd, virt_addr);
	if(p4d_none(*p4d) || p4d_bad(*p4d)){
		pr_info("[CHAR_DEV] Page not present\n");
		return 0;
	}
	
	pud = pud_offset(p4d, virt_addr);
	if(pud_none(*pud) || pud_bad(*pud)){
		pr_info("[CHAR_DEV] Page not present\n");
		return 0;
	}

	pmd = pmd_offset(pud, virt_addr);
	if(pmd_none(*pmd) || pmd_bad(*pmd)){
		pr_info("[CHAR_DEV] Page not present\n");
		return 0;
	}

	pte = pte_offset_kernel(pmd, virt_addr);
	if(pte_none(*pte)){
		pr_info("[CHAR_DEV] Page not present\n");
		return 0;
	}

	phy_addr = (pte_pfn(*pte) << PAGE_SHIFT) | (virt_addr & ~PAGE_MASK);
	return phy_addr;
}

static long device_ioctl(struct file* file, unsigned int ioctl_num, unsigned long ioctl_param)
{
	struct mm_struct* mm = current->mm;
	struct query_arg_t unused_arg;
	struct query_arg_t *arg = &unused_arg;
	if (copy_from_user(arg,(struct query_arg_t *)ioctl_param,sizeof(struct query_arg_t)))
		return -EACCES;

	switch (ioctl_num)
	{
		case IOCTL_PHYS_TO_VIRT:
			// Convert physical address to virtual address
			unsigned long virt_addr = arg->virt_addr;
			unsigned long phy_addr = page_table_walk(virt_addr, mm);
			arg -> phys_addr = phy_addr;

			if (copy_to_user((struct query_arg_t *)ioctl_param,arg,sizeof(struct query_arg_t)))
				return -EACCES;

			break;
		case IOCTL_WRITE_TO_PHYS:
			
		
			*(char*)phys_to_virt(arg->phys_addr) = arg->byte_val;

			// Alternative - 1
			// void* vaddr = (void *)((unsigned long)kmap_local_pfn(arg->phys_addr >> PAGE_SHIFT) | (arg->phys_addr & ~PAGE_MASK));
			// *(char*)vaddr = arg->byte_val;
			// kunmap_local(vaddr);


			// Alternative - 2
			// struct page* page = pfn_to_page(arg->phys_addr >> PAGE_SHIFT);
			// void* vaddr = (void *)((unsigned long)kmap(page) | (arg->phys_addr & ~PAGE_MASK));
			// *(char*)vaddr = arg->byte_val;
			// kunmap(page);
			
			break;
	}
	return 0;
}

struct file_operations fops = {
	.open = NULL,
	.release = NULL,
	.read = NULL,
	.write = NULL,
	.unlocked_ioctl = device_ioctl
};

static int __init register_device(void)
{
	int ret;
    struct device *dev_ret;
 
 
    if ((ret = alloc_chrdev_region(&dev, FIRST_MINOR, MINOR_CNT, "query_ioctl")) < 0)
    {
        return ret;
    }
 
    cdev_init(&c_dev, &fops);
 
    if ((ret = cdev_add(&c_dev, dev, MINOR_CNT)) < 0)
    {
        return ret;
    }

    if (IS_ERR(cl = class_create(THIS_MODULE, "addr_translate")))
    {
        cdev_del(&c_dev);
        unregister_chrdev_region(dev, MINOR_CNT);
        return PTR_ERR(cl);
    }
    if (IS_ERR(dev_ret = device_create(cl, NULL, dev, NULL, DEVICE_NAME)))
    {
        class_destroy(cl);
        cdev_del(&c_dev);
        unregister_chrdev_region(dev, MINOR_CNT);
        return PTR_ERR(dev_ret);
    }
 
    return 0;
}

static void __exit dereg_device(void)
{
	device_destroy(cl, dev);
    class_destroy(cl);
    cdev_del(&c_dev);
    unregister_chrdev_region(dev, MINOR_CNT);
}

module_init(register_device);
module_exit(dereg_device);

MODULE_DESCRIPTION("2_I");
MODULE_AUTHOR("CS695");
MODULE_LICENSE("GPL");
